/************************************************************************************
 * Copyright (c) 2012 Paul Lawitzki
 * Copyright (c) 2014 Jose Collas : Wrapper code
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
 * OR OTHER DEALINGS IN THE SOFTWARE.
 ************************************************************************************/

package com.example.jairo.wristcare.util;

import android.hardware.SensorEvent;
import android.hardware.SensorManager;

import java.util.Timer;
import java.util.TimerTask;

/**
 * Provides a wrapper around code that fuses sensor data.
 * @author 2012 Paul Lawitzki
 * @author 2014 Jose Collas
 *
 * @version 0.7 1/26/14.
 *
 */
public class SensorFusion {

    // Velocidades angulares del giroscopio
    private float[] gyro = new float[3];

    // Matriz de rotacion de los datos del giroscopio
    private float[] gyroMatrix = new float[9];

    // Ángulos de orientación de la matriz del gyroscopio
    private float[] gyroOrientation = new float[3];

    // Vector del campo magnético
    private float[] magnet = new float[3];

    // Vector del acelerómetro
    private float[] accel = new float[3];

    // Ángulos de orientación del acelerómetro y del magnetómetro
    private float[] accMagOrientation = new float[3];

    // Ángulos de orientación finales de la fusión de sensores
    private float[] fusedOrientation = new float[3];

    private float[] selectedOrientation = fusedOrientation;

    public enum Mode {
        ACC_MAG, GYRO, FUSION
    }

    // Matriz de rotación basada en el acelerómetro y magnetómetro
    private float[] rotationMatrix = new float[9];

    public static final float EPSILON = 0.000000001f;
    private static final float NS2S = 1.0f / 1000000000.0f;
    private long timestamp;
    private boolean initState = true;

    public static final int TIME_CONSTANT = 30;
    public static final float FILTER_COEFFICIENT = 0.98f;
    private Timer fuseTimer = new Timer();

    public SensorFusion() {

        gyroOrientation[0] = 0.0f;
        gyroOrientation[1] = 0.0f;
        gyroOrientation[2] = 0.0f;

        // Inicializar la matriz del giroscopio con la matriz identidad
        gyroMatrix[0] = 1.0f;
        gyroMatrix[1] = 0.0f;
        gyroMatrix[2] = 0.0f;
        gyroMatrix[3] = 0.0f;
        gyroMatrix[4] = 1.0f;
        gyroMatrix[5] = 0.0f;
        gyroMatrix[6] = 0.0f;
        gyroMatrix[7] = 0.0f;
        gyroMatrix[8] = 1.0f;

        // Esperar un segundo hasta que los datos del giroscopio y el magnetómetro/acelerómetro
        // se inicialicen. Después se programará la tarea de filtrado complementario

        fuseTimer.scheduleAtFixedRate(new calculateFusedOrientationTask(),
                1000, TIME_CONSTANT);

    }

    public double getAzimuth() {
        return selectedOrientation[0] * 180 / Math.PI;
    }

    public double getPitch() {
        return selectedOrientation[1] * 180 / Math.PI;
    }

    public double getRoll() {
        return selectedOrientation[2] * 180 / Math.PI;
    }

    public void setMagnet(float[] sensorValues) {
        System.arraycopy(sensorValues, 0, magnet, 0, 3);
    }

    public void setAccel(float[] sensorValues) {
        System.arraycopy(sensorValues, 0, accel, 0, 3);

    }

    public  void setMode(Mode mode) {
        switch (mode) {
            case ACC_MAG:
                selectedOrientation = accMagOrientation;
                break;
            case GYRO:
                selectedOrientation = gyroOrientation;
                break;
            case FUSION:
                selectedOrientation = fusedOrientation;
                break;
            default:
                selectedOrientation = fusedOrientation;
                break;
        }
    }

    // Calcula los ángulos de orientación de las salidas del acelerómetro y del magnetómetro
    public void calculateAccMagOrientation() {
        if (SensorManager.getRotationMatrix(rotationMatrix, null, accel, magnet)) {
            SensorManager.getOrientation(rotationMatrix, accMagOrientation);
        }
    }

    // Esta función se coge de la referencia de Android en:
    // http://developer.android.com/reference/android/hardware/SensorEvent.html#values
    // Calcula un vector de rotación a partir de los valores de la velocidad angular del giroscopio

    private void getRotationVectorFromGyro(float[] gyroValues,
                                           float[] deltaRotationVector,
                                           float timeFactor) {
        float[] normValues = new float[3];

        // Calcula la velocidad angular de la muestra
        float omegaMagnitude =
                (float) Math.sqrt(gyroValues[0] * gyroValues[0] +
                        gyroValues[1] * gyroValues[1] +
                        gyroValues[2] * gyroValues[2]);

        // Normaliza el vector de rotación si es lo suficientemente grande como para llegar al eje
        if (omegaMagnitude > EPSILON) {
            normValues[0] = gyroValues[0] / omegaMagnitude;
            normValues[1] = gyroValues[1] / omegaMagnitude;
            normValues[2] = gyroValues[2] / omegaMagnitude;
        }

        // Integra alrededor de este eje con la velocidad angular en el lapso
        // con el fin de obtener una rotación delta de esta muestra durante el lapso
        // Se convertirá esta representación ángulo-eje de la rotación delta
        // en un quaternión antes de pasarlo a una matriz de rotación.
        float thetaOverTwo = omegaMagnitude * timeFactor;
        float sinThetaOverTwo = (float) Math.sin(thetaOverTwo);
        float cosThetaOverTwo = (float) Math.cos(thetaOverTwo);
        deltaRotationVector[0] = sinThetaOverTwo * normValues[0];
        deltaRotationVector[1] = sinThetaOverTwo * normValues[1];
        deltaRotationVector[2] = sinThetaOverTwo * normValues[2];
        deltaRotationVector[3] = cosThetaOverTwo;
    }

    // Esta función realiza la integración de los datos del giroscopio.
    // Escribe la orientación basada en el giroscopio en gyroOrientation.
    public void gyroFunction(SensorEvent event) {
        // No empezar hasta que la primera orientación se haya conseguido
        if (accMagOrientation == null)
            return;

        // inicialización de la matriz de rotación basada en el giroscopio
        if (initState) {
            float[] initMatrix;
            initMatrix = getRotationMatrixFromOrientation(accMagOrientation);
            float[] test = new float[3];
            SensorManager.getOrientation(initMatrix, test);
            gyroMatrix = matrixMultiplication(gyroMatrix, initMatrix);
            initState = false;
        }

        // copia de los nuevos valores del giroscopio en el array
        // convierte los datos en crudo en un vector de rotación
        float[] deltaVector = new float[4];
        if (timestamp != 0) {
            final float dT = (event.timestamp - timestamp) * NS2S;
            System.arraycopy(event.values, 0, gyro, 0, 3);
            getRotationVectorFromGyro(gyro, deltaVector, dT / 2.0f);
        }

        // con la medida hecha, guarda el tiempo actual para el siguiente intervalo
        timestamp = event.timestamp;

        // convierte el vector de rotación en una matriz de rotación
        float[] deltaMatrix = new float[9];
        SensorManager.getRotationMatrixFromVector(deltaMatrix, deltaVector);

        // aplicar el nuevo intervalo de rotación en la matriz de rotación basada en el giroscopio
        gyroMatrix = matrixMultiplication(gyroMatrix, deltaMatrix);

        // conseguir la orientación basada en el giroscopio de la matriz de rotación
        SensorManager.getOrientation(gyroMatrix, gyroOrientation);
    }

    private float[] getRotationMatrixFromOrientation(float[] o) {
        float[] xM = new float[9];
        float[] yM = new float[9];
        float[] zM = new float[9];

        float sinX = (float) Math.sin(o[1]);
        float cosX = (float) Math.cos(o[1]);
        float sinY = (float) Math.sin(o[2]);
        float cosY = (float) Math.cos(o[2]);
        float sinZ = (float) Math.sin(o[0]);
        float cosZ = (float) Math.cos(o[0]);

        // rotación sobre el eje x (pitch)
        xM[0] = 1.0f;xM[1] = 0.0f;xM[2] = 0.0f;
        xM[3] = 0.0f;xM[4] = cosX;xM[5] = sinX;
        xM[6] = 0.0f;xM[7] = -sinX;xM[8] = cosX;

        // rotación sobre el eje y (roll)
        yM[0] = cosY;yM[1] = 0.0f;yM[2] = sinY;
        yM[3] = 0.0f;yM[4] = 1.0f;yM[5] = 0.0f;
        yM[6] = -sinY;yM[7] = 0.0f;yM[8] = cosY;

        // rotación sobre el eje z (azimuth)
        zM[0] = cosZ;zM[1] = sinZ;zM[2] = 0.0f;
        zM[3] = -sinZ;zM[4] = cosZ;zM[5] = 0.0f;
        zM[6] = 0.0f;zM[7] = 0.0f;zM[8] = 1.0f;

        // El orden de rotación es y, x, z (roll, pitch, azimuth)
        float[] resultMatrix = matrixMultiplication(xM, yM);
        resultMatrix = matrixMultiplication(zM, resultMatrix);
        return resultMatrix;
    }

    private float[] matrixMultiplication(float[] A, float[] B) {
        float[] result = new float[9];

        result[0] = A[0] * B[0] + A[1] * B[3] + A[2] * B[6];
        result[1] = A[0] * B[1] + A[1] * B[4] + A[2] * B[7];
        result[2] = A[0] * B[2] + A[1] * B[5] + A[2] * B[8];

        result[3] = A[3] * B[0] + A[4] * B[3] + A[5] * B[6];
        result[4] = A[3] * B[1] + A[4] * B[4] + A[5] * B[7];
        result[5] = A[3] * B[2] + A[4] * B[5] + A[5] * B[8];

        result[6] = A[6] * B[0] + A[7] * B[3] + A[8] * B[6];
        result[7] = A[6] * B[1] + A[7] * B[4] + A[8] * B[7];
        result[8] = A[6] * B[2] + A[7] * B[5] + A[8] * B[8];

        return result;
    }

    class calculateFusedOrientationTask extends TimerTask {
        public void run() {
            float oneMinusCoeff = 1.0f - FILTER_COEFFICIENT;

            /*
             * Arreglo para el problema de transición para 179∞ <--> -179∞
             * Comprueba si uno de los dos ángulos de orientación (giroscopio o acclerómetro/magnetómetro) es negativo mientras que el otro es positivo.
             * Si es así, añadir 360∞ (2 * math.PI) al valor negativo, realizar la fusion de sensores, y eliminar el 360∞ del resultado
             * si esto es mayor que 180∞. Esto estabiliza la salida en los casos de transición-de-positivo-a-negativo.
             */

            // azimuth
            if (gyroOrientation[0] < -0.5 * Math.PI && accMagOrientation[0] > 0.0) {
                fusedOrientation[0] = (float) (FILTER_COEFFICIENT * (gyroOrientation[0] + 2.0 * Math.PI) + oneMinusCoeff * accMagOrientation[0]);
                fusedOrientation[0] -= (fusedOrientation[0] > Math.PI) ? 2.0 * Math.PI : 0;
            } else if (accMagOrientation[0] < -0.5 * Math.PI && gyroOrientation[0] > 0.0) {
                fusedOrientation[0] = (float) (FILTER_COEFFICIENT * gyroOrientation[0] + oneMinusCoeff * (accMagOrientation[0] + 2.0 * Math.PI));
                fusedOrientation[0] -= (fusedOrientation[0] > Math.PI) ? 2.0 * Math.PI : 0;
            } else {
                fusedOrientation[0] = FILTER_COEFFICIENT * gyroOrientation[0] + oneMinusCoeff * accMagOrientation[0];
            }

            // pitch
            if (gyroOrientation[1] < -0.5 * Math.PI && accMagOrientation[1] > 0.0) {
                fusedOrientation[1] = (float) (FILTER_COEFFICIENT * (gyroOrientation[1] + 2.0 * Math.PI) + oneMinusCoeff * accMagOrientation[1]);
                fusedOrientation[1] -= (fusedOrientation[1] > Math.PI) ? 2.0 * Math.PI : 0;
            } else if (accMagOrientation[1] < -0.5 * Math.PI && gyroOrientation[1] > 0.0) {
                fusedOrientation[1] = (float) (FILTER_COEFFICIENT * gyroOrientation[1] + oneMinusCoeff * (accMagOrientation[1] + 2.0 * Math.PI));
                fusedOrientation[1] -= (fusedOrientation[1] > Math.PI) ? 2.0 * Math.PI : 0;
            } else {
                fusedOrientation[1] = FILTER_COEFFICIENT * gyroOrientation[1] + oneMinusCoeff * accMagOrientation[1];
            }

            // roll
            if (gyroOrientation[2] < -0.5 * Math.PI && accMagOrientation[2] > 0.0) {
                fusedOrientation[2] = (float) (FILTER_COEFFICIENT * (gyroOrientation[2] + 2.0 * Math.PI) + oneMinusCoeff * accMagOrientation[2]);
                fusedOrientation[2] -= (fusedOrientation[2] > Math.PI) ? 2.0 * Math.PI : 0;
            } else if (accMagOrientation[2] < -0.5 * Math.PI && gyroOrientation[2] > 0.0) {
                fusedOrientation[2] = (float) (FILTER_COEFFICIENT * gyroOrientation[2] + oneMinusCoeff * (accMagOrientation[2] + 2.0 * Math.PI));
                fusedOrientation[2] -= (fusedOrientation[2] > Math.PI) ? 2.0 * Math.PI : 0;
            } else {
                fusedOrientation[2] = FILTER_COEFFICIENT * gyroOrientation[2] + oneMinusCoeff * accMagOrientation[2];
            }

            // sobreescribir la matriz del giroscopio y de orientación con fusión de orientación para compensar el gyro drift.
            gyroMatrix = getRotationMatrixFromOrientation(fusedOrientation);
            System.arraycopy(fusedOrientation, 0, gyroOrientation, 0, 3);


            // update sensor output in GUI
            //mainHandler.post(updateOreintationDisplayTask);
        }
    }


}